#!/usr/bin/env python3
"""
Train function using various of classifiers
TODO:
DONE    1. train,valid,test accuracy
DONE    2. canonicalization: It looks that canonicalization does not work well.
DONE    3. SGD with learning rate decreasing
DONE    4. Dropout and regularizations
        5. PEPS 2x3 for MNIST, further parametrization for the MPO
DONE    6. Use args to give dimensions for the MLP
DONE    7. Data augmentation
        8. Resblock implementation of MPS factors
        9. Checking SVD, looking further for low-rank structures
"""

import time
import os
import sys

import numpy as np
import torch
import torchvision
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms
from tncom.args import args
from tncom import conv
from tncom import mpofc
from tncom import mpomlp
from tncom import opmlp
from tncom.utils import mylog
import tncom.pepo as pepo
from tncom.mpolinear import MPOLinear3

def train(args, model, train_loader, optimizer, epoch):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        if not args.ingpu:
            data, target = data.to(args.device), target.to(args.device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
    return loss.item()

def test(args, model, valid_loader,test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    valid_loss=0
    valid_correct=0
    n_test = 0
    n_valid = 0
    with torch.no_grad():
        for data, target in test_loader:
            if not args.ingpu:
                data, target = data.to(args.device), target.to(args.device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item() # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True) # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()
            n_test += args.test_batch_size

    with torch.no_grad():
        for data, target in valid_loader:
            if not args.ingpu:
                data, target = data.to(args.device), target.to(args.device)
            output = model(data)
            valid_loss += F.nll_loss(output, target, reduction='sum').item() # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True) # get the index of the max log-probability
            valid_correct += pred.eq(target.view_as(pred)).sum().item()
            n_valid += args.test_batch_size
    test_loss /= n_test
    correct /= n_test
    valid_loss /= n_valid
    valid_correct /= n_valid
    return valid_loss,valid_correct,test_loss,correct

def main():
    start_time = time.time()
    torch.manual_seed(args.seed)
    if args.log:
        with open(args.log, 'a', newline='\n') as f:
            f.write("[ "+" ".join(sys.argv)+u' ]\n')
            f.write(str(args)+u'\n')
    if  args.net == 'mpofc':
        dims = [int(i) for i in args.dims.split(",")]
        assert( len(dims) == 4)
        mylog("MPO dims "+str(dims)+" D= %d"%(np.prod(dims)))
    if args.net == 'conv':
        net = conv.Net(args)
    if args.net == 'conv2':
        net = conv.Net2(args)
    elif args.net == 'fc' or args.net == 'mpofc' or args.net == 'lrfc' or args.net == 'mpofc3':
        net = mpofc.Net(args)
    elif args.net == 'mpomlp':
        net = mpomlp.Net(args)
    elif args.net == 'opmlp':
        net = opmlp.Net(args)
    elif args.net == 'pepo':
        net = pepo.Net(args)
        print("PEPO")
    else:
        raise ValueError('Unknown net: {}'.format(args.net))
    net_time = time.time()

    params = list(net.parameters())
    params = list(filter(lambda p: p.requires_grad, params))
    nparams = int(sum([np.prod(p.shape) for p in params]))
    mylog('Total number of trainable parameters: {}'.format(nparams))
    named_params = list(net.named_parameters())

    sys.stdout.write("copying net to GPU...");sys.stdout.flush()
    net.to(args.device)
    mylog('{}\n'.format(net))
    print("done. %.4f"%(time.time()-net_time))

    if args.optimizer == 'sgd':
        optimizer = torch.optim.SGD(params, lr=args.lr)
    elif args.optimizer == 'sgdm':
        optimizer = torch.optim.SGD(params, lr=args.lr, momentum=0.9)
    elif args.optimizer == 'rmsprop':
        optimizer = torch.optim.RMSprop(params, lr=args.lr, alpha=0.99)
    elif args.optimizer == 'adam':
        optimizer = torch.optim.Adam(params, lr=args.lr, betas=(0.9, 0.999),weight_decay=args.l2)
    elif args.optimizer == 'adam0.5':
        optimizer = torch.optim.Adam(params, lr=args.lr, betas=(0.5, 0.999))
    else:
        raise ValueError('Unknown optimizer: {}'.format(args.optimizer))

    init_time = time.time() - start_time
    mylog('init_time = {:.3f} Sec.'.format(init_time))
    kwargs = {'num_workers': 8, 'pin_memory': True} if args.use_cuda else {}
    sys.stdout.write("Loading data...")
    sys.stdout.flush()
    init_time=time.time()
    print("Dataset:",args.dataset)
    if args.dataset == 'mnist':
        data = datasets.MNIST
        data_dir = '../data'
        transform = transforms.Compose([ transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,)) ])
    elif args.dataset == 'fmnist':
        data = datasets.FashionMNIST
        data_dir = '../data_fashionMNIST'
        transform = transforms.Compose([ transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,)) ])
    elif args.dataset == 'cifar10':
        data = datasets.CIFAR10
        data_dir = '../data_cifar10'
        transform = transforms.Compose( [transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])

    train_loader = torch.utils.data.DataLoader( data(data_dir, train=True, download=True, transform=transform), batch_size=args.batch_size, shuffle=False, **kwargs)
    test_loader = torch.utils.data.DataLoader(data(data_dir, train=False, transform=transform), batch_size=args.test_batch_size, shuffle=False, **kwargs)


    num_train = 50000 if args.dataset == 'cifar10' else 60000
    valid_size= args.valid
    indices = list(range(num_train))
    split = valid_size

    np.random.seed(args.seed)
    np.random.shuffle(indices)

    train_idx, valid_idx = indices[split:], indices[:split]

    train_sampler = torch.utils.data.sampler.SubsetRandomSampler(train_idx)
    valid_sampler = torch.utils.data.sampler.SubsetRandomSampler(valid_idx)

    train_loader = torch.utils.data.DataLoader( data(data_dir, train=True, download=True, transform=transform), batch_size=args.batch_size, sampler=train_sampler, shuffle=False, **kwargs)
    valid_loader = torch.utils.data.DataLoader( data(data_dir, train=True, download=True, transform=transform), batch_size=args.test_batch_size, sampler=valid_sampler, shuffle=False, **kwargs)

    if args.aug:
        print("data augmentation")
        train_transform0 = transforms
        train_loader0 = torch.utils.data.DataLoader( data(data_dir, train=True, download=True, transform=train_transform0), batch_size=args.batch_size, sampler=train_sampler, shuffle=False, **kwargs)
        train_transform = transforms.Compose([ transforms.RandomHorizontalFlip(p=0.5), transforms.RandomRotation(30), transforms.ToTensor(), transform ])
        train_loader = torch.utils.data.DataLoader( data(data_dir, train=True, download=True, transform=train_transform), batch_size=args.batch_size, sampler=train_sampler, shuffle=False, **kwargs)

    print(" done %.3f seconds."%(time.time()-init_time))
    train_data=[]
    test_data=[]
    valid_data=[]
    copy_time = time.time()
    if args.ingpu and args.use_cuda:
        sys.stdout.write("Copying Data to GPU...")
    else:
        sys.stdout.write("Data in CPU...")
    sys.stdout.flush()
    init_time=time.time()
    if args.ingpu and args.use_cuda:
        for batch_idx, (data, target) in enumerate(train_loader):
            data, target = data.to(args.device), target.to(args.device)
            train_data.append((data,target))
        for batch_idx, (data, target) in enumerate(valid_loader):
            data, target = data.to(args.device), target.to(args.device)
            valid_data.append((data,target))
        for batch_idx, (data, target) in enumerate(test_loader):
            data, target = data.to(args.device), target.to(args.device)
            test_data.append((data,target))
    else:
        train_data = train_loader
        valid_data = valid_loader
        test_data = test_loader
    print(" done %.3f seconds."%(time.time()-copy_time))
    print("Train data",len(train_data),"batches", " %.3f seconds."%(time.time()-init_time))
    print("Training...")
    best_result = -10.0
    for epoch in range(1, args.epochs + 1):
        if args.cano > 0 and epoch >0:
            if epoch % args.cano == 0:
                for idx, m in enumerate(net.modules()):
                    if isinstance(m,MPOLinear3) and idx == 1: # the first MPOLinear layer
                        m.cano()
        lr = args.lr * (args.lr_decay ** (epoch // args.lr_decay_epochs))
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr
        t1=time.time()
        t0_train = time.time()
        loss=train(args, net, train_data, optimizer, epoch)
        if args.aug2:
            loss=train(args, net, train_loader0, optimizer, epoch)
        t_train = time.time()-t0_train
        t0_test = time.time()
        loss_valid,accu_valid,loss_test,accu_test=test(args, net, valid_data,test_data)
        t_test = time.time()-t0_test
        save_flag = ''
        if accu_test > best_result:
            best_result = accu_test
            if args.save:
                torch.save([args,net,test_data],args.save)
                save_flag = 'saved'
        if args.log_interval < 0: args.log_interval = args.epochs
        bin=int(args.epochs/args.log_interval)
        if bin <1: bin=1
        if(epoch %  bin == 0 or epoch == args.epochs):
            mylog("#%d lr=%g loss= ( %.3g, %.3g, %.3g),\taccuracy= (%.4f, %.4f) best=%.4f\t%.2f %.2f Sec. %s"%(epoch,lr,loss,loss_valid,loss_test,accu_valid,accu_test,best_result,t_train,t_test,save_flag))


if __name__ == '__main__':
    main()
