import math
import sys
import numpy as np

import torch
from torch.nn.parameter import Parameter
from torch.nn import functional as F
from torch.nn import init
from torch.nn.modules.module import Module
from .utils import AttrProxy

class MPOLinear2(Module):
    __constants__ = ['bias']
    def __init__(self, Din, Dout, bias=False, chi=2):
        """
        Din (and Dout) should be a tuple containing all input (output) dimensions
        """
        super(MPOLinear2, self).__init__()
        self.Din=Din
        self.Dout=Dout
        self.bondim=[chi for i in Din]
        self.bondim[-1]=1
        print("Din=",Din)
        print("Dout=",Dout)
        assert(len(self.Din) == len(self.Dout))
        if bias:
            self.bias = Parameter(torch.zeros(Dout))
        else:
            self.register_parameter('bias', None)
        self.tensors=[]
        npin=1
        npout=1
        for i,din in enumerate(self.Din):
            npin *= din
            npout *= self.Dout[i]
        for i,din in enumerate(self.Din):
            dout=self.Dout[i]
            a=torch.rand(self.bondim[i-1],self.bondim[i],dout,din)/math.sqrt(npout)
            exec("self.tensors_"+str(i)+"=Parameter(a.clone())")
        self.tensors = AttrProxy(self, 'tensors_')

        print(self)
        print("Parameters in the class")
        params = list(self.parameters())
        params = list(filter(lambda p: p.requires_grad, params))
        nparams = int(sum([np.prod(p.shape) for p in params]))
        print('Total number of trainable parameters: {}'.format(nparams))
        for param in self.parameters():
            print(type(param.data), param.size())


    def forward(self, input):
        for i,din in enumerate(self.Din):
            if(i==0):
                tensor=self.tensors[i]
            else:
                tensor = ncon([tensor,self.tensors[i]],([-1,1,-2,-3],[1,-4,-5,-6]),[1],[-1,-4,-2,-5,-3,-6]).contiguous().view(tensor.shape[0],self.tensors[i].shape[1],tensor.shape[2]*self.tensors[i].shape[2],tensor.shape[3]*self.tensors[i].shape[3])
        tensor = tensor.contiguous().view(tensor.shape[2],tensor.shape[3])
        return (tensor@input.t()).t()


class LRLinear(Module):
    __constants__ = ['bias']

    def __init__(self, in_features, out_features, bias=True, chi=10):
        super(LRLinear, self).__init__()
        D=chi
        self.in_features = in_features
        self.out_features = out_features
        self.bias_flag=False
        if bias:
            self.bias_flag=True
            self.bias = Parameter(torch.zeros(out_features))
            self.bias.data = self.bias.data-self.bias.data
        else:
            self.register_parameter('bias', None)
        self.w1=Parameter(torch.rand(in_features,D)*0.0001)
        self.w2=Parameter(torch.rand(D,out_features)*0.0001)

    def forward(self, input):
        if self.bias_flag:
            return input@self.w1@self.w2+self.bias
        else:
            return input@self.w1@self.w2

    def extra_repr(self):
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )

class MPOLinear3(Module):
    """
    Using general ball-MPO contraction, hence has lower computational complexity than MPOLinear2.
    Add bias
    """
    __constants__ = ['bias']
    def __init__(self, Din, Dout, bias=False, chi=2,seed = -1):
        """
        Din (and Dout) should be a tuple containing all input (output) dimensions
        """
        super(MPOLinear3, self).__init__()
        self.Din=Din
        self.Dout=Dout
        self.bondim=[chi for i in Din]
        self.bondim[-1]=1
        print("Din=",Din,"Dout=",Dout)
        assert(len(self.Din) == len(self.Dout))
        self.tensors=[]
        self.npin = np.prod(self.Din)
        self.npout = np.prod(self.Dout)
        if seed > 0:
            torch.manual_seed(seed)
        for i,din in enumerate(self.Din):
            dout=self.Dout[i]
            a=torch.rand(self.bondim[i-1],self.bondim[i],din,dout)/math.sqrt(self.npout)
            exec("self.tensors_"+str(i)+"=Parameter(a.clone())")
        if bias:
            self.bias = Parameter(torch.zeros([self.npout,1]))
        else:
            self.register_parameter('bias', None)
        self.tensors = AttrProxy(self, 'tensors_')

        print(self)
        print("Parameters in the class")
        params = list(self.parameters())
        params = list(filter(lambda p: p.requires_grad, params))
        nparams = int(sum([np.prod(p.shape) for p in params]))
        print('Total number of trainable parameters: {}'.format(nparams))
        for param in self.parameters():
            print(type(param.data), param.size())

    def cano(self):
        """ Canonicalization of the MPO """
        sys.stdout.write("CANO "+str([self.tensors[i].shape[1] for i in range(len(self.Din))])+" ");sys.stdout.flush()
        for i in range(len(self.Din)):
            D = self.tensors[i].shape[1]
            if i>0:
                mat = torch.einsum('ab,bcde->acde',R,self.tensors[i])
                shape0 = R.shape[0]
            else:
                mat = self.tensors[i]
                shape0 = mat.shape[0]
            if i<len(self.Din)-1:
                mat = mat.permute([0,2,3,1]).reshape([-1,D])
                Q,R = torch.qr(mat)
                newtensor = Q.reshape([shape0,self.tensors[i].shape[2],self.tensors[i].shape[3],-1]).permute([0,3,1,2])
            else:
                newtensor = mat
            exec("self.tensors_"+str(i)+".data=newtensor")
        print("-->",[self.tensors[i].shape[1] for i in range(len(self.Din))])

    def dmrg(self, cutoff):
        """ Rounding using DMRG with a given cutoff  """
        print("Before shape:",[self.tensors[i].shape[1] for i in range(len(self.Din))])
        print("cutoff=",cutoff)
        self.cano() # first do canonicalization to the left canonical form
        error = 0
        for i in range(len(self.Din)-1,0,-1):
            D = self.tensors[i-1].shape[1]
            shape_up = self.tensors[i-1].shape
            shape_down = self.tensors[i].shape
            mat_up = self.tensors[i-1].permute([0,2,3,1]).reshape([-1,D])
            mat_down = self.tensors[i].reshape([D,-1])
            Q0,R0 = torch.qr(mat_up)
            Q1,R1 = torch.qr(mat_down.t())
            mat = R0 @ R1.t()
            [U,s,V] = torch.svd(mat)
            print("D=",D,"s=",s/s.max())
            s_eff = s[s/s.max()>cutoff]
            myd = min(len(s_eff),D)
            error = s[myd:].sum()
            U=U[:,:myd]
            V=V[:,:myd]
            s_eff=s_eff[:myd]
            S=torch.diag(s_eff)
            U = (Q0@U@S).reshape(shape_up[0],shape_up[2],shape_up[3],-1).permute([0,3,1,2])
            V = ((V.t())@Q1.t()).reshape(-1,shape_down[1],shape_down[2],shape_down[3])
            exec("self.tensors_"+str(i-1)+".data=U")
            exec("self.tensors_"+str(i)+".data=V")
        print("after shape:",[self.tensors[i].shape[1] for i in range(len(self.Din))],"error=",error)

    def forward(self, input):
        input = input.reshape(input.shape[0],1,1,self.Din[0],-1)
        for i in range(len(self.Din)):
            input = torch.einsum("bijkl,jakm->bimal",input,self.tensors[i])
            Dnext = self.Din[i+1] if i<len(self.Din)-1 else 1
            newshape=[input.shape[0],input.shape[1]*input.shape[2],input.shape[3],Dnext,-1]
            input = input.contiguous().view(newshape)
        return input.contiguous().view(input.shape[0],-1)
        
class OperatorLinear(Module):
    """
    Use sequence of operators to span a large space from the input space with a smaller dimension
    """
    __constants__ = ['bias']
    def __init__(self, Din, Dout, bias=False, chi=2):
        """
        Din (and Dout) should be a tuple containing all input (output) dimensions
        """
        super(OperatorLinear, self).__init__()
        self.Din=Din
        self.Dout=Dout
        self.bondim=[chi for i in Din]
        self.bondim[-1]=1
        print("Din=",Din,"Dout=",Dout)
        assert(len(self.Din) == len(self.Dout))
        self.tensors=[]
        self.npin = np.prod(self.Din)
        self.npout = np.prod(self.Dout)
        for i in range(len(self.Din)-1):
            d0 = self.Din[i] if i==0 else self.Dout[i]
            d1 = self.Din[i+1]
            d2 = self.Dout[i]
            d3 = self.Dout[i+1]
            a=torch.rand(d0,d1,d2,d3)/math.sqrt(self.npout)
            exec("self.operators_"+str(i)+"=Parameter(a.clone())")
        self.operators = AttrProxy(self, 'operators_')
        params = list(self.parameters())
        params = list(filter(lambda p: p.requires_grad, params))
        nparams = int(sum([np.prod(p.shape) for p in params]))
        print('Operator layer: total number of trainable parameters: {}'.format(nparams))
        for param in self.parameters():
            print(type(param.data), param.size())

    def forward(self, input):
        input = input.reshape(input.shape[0],1,self.Din[0],self.Din[1],-1)
        for i in range(len(self.Din)-1):
            input = torch.einsum("bxijk,ijlm->bxlmk",input,self.operators[i])
            Dnext = 1 if i==len(self.Din)-2 else self.operators[i+1].shape[1]
            newshape=[input.shape[0],input.shape[1]*input.shape[2],input.shape[3],Dnext,-1]
            input = input.contiguous().view(newshape)
        return input.contiguous().view(input.shape[0],-1)
